/*
 * SNMPStats Module
 * Copyright (C) 2006 SOMA Networks, INC.
 * Written by: Jeffrey Magder (jmagder@somanetworks.com)
 *
 * This file is part of opensips, a free SIP server.
 *
 * opensips is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * opensips is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
 * USA
 *
 * History:
 * --------
 * 2006-11-23 initial version (jmagder)
 *
 * Note: this file originally auto-generated by mib2c using
 *        : mib2c.scalar.conf,v 1.9 2005/01/07 09:37:18 dts12 Exp $
 *
 * This file defines all registration and handling of all scalars defined in the
 * OPENSER-SIP-COMMON-MIB.  Please see OPENSER-SIP-COMMON-MIB for the complete
 * descriptions of the individual scalars.
 */
#ifndef OPENSERSIPCOMMONOBJECTS_H
#define OPENSERSIPCOMMONOBJECTS_H

void init_openserSIPCommonObjects(void);

Netsnmp_Node_Handler handle_openserSIPProtocolVersion;
Netsnmp_Node_Handler handle_openserSIPServiceStartTime;
Netsnmp_Node_Handler handle_openserSIPEntityType;
Netsnmp_Node_Handler handle_openserSIPSummaryInRequests;
Netsnmp_Node_Handler handle_openserSIPSummaryOutRequests;
Netsnmp_Node_Handler handle_openserSIPSummaryInResponses;
Netsnmp_Node_Handler handle_openserSIPSummaryOutResponses;
Netsnmp_Node_Handler handle_openserSIPSummaryTotalTransactions;
Netsnmp_Node_Handler handle_openserSIPCurrentTransactions;
Netsnmp_Node_Handler handle_openserSIPNumUnsupportedUris;
Netsnmp_Node_Handler handle_openserSIPNumUnsupportedMethods;
Netsnmp_Node_Handler handle_openserSIPOtherwiseDiscardedMsgs;

#endif /* OPENSERSIPCOMMONOBJECTS_H */
