/*
 * This file is automatically generated, do not edit
 *
 * Copyright (C) 2001-2003 FhG Fokus
 *
 * This file is part of opensips, a free SIP server.
 *
 * opensips is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * opensips is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */


#ifndef DIGEST_KEYS_H
#define DIGEST_KEYS_H

#define _user_ 0x72657375   /* "user" */
#define _name_ 0x656d616e   /* "name" */
#define _real_ 0x6c616572   /* "real" */
#define _nonc_ 0x636e6f6e   /* "nonc" */
#define _resp_ 0x70736572   /* "resp" */
#define _onse_ 0x65736e6f   /* "onse" */
#define _cnon_ 0x6e6f6e63   /* "cnon" */
#define _opaq_ 0x7161706f   /* "opaq" */
#define _algo_ 0x6f676c61   /* "algo" */
#define _rith_ 0x68746972   /* "rith" */

#endif /* DIGEST_KEYS_H */
