/**
 * Autogenerated by Thrift
 *
 * DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
 */
#include "cassandra_constants.h"

namespace org { namespace apache { namespace cassandra {

const cassandraConstants g_cassandra_constants;

cassandraConstants::cassandraConstants() {
  VERSION = "19.19.0";

}

}}} // namespace

