/*
 * opensips osp module.
 *
 * This module enables opensips to communicate with an Open Settlement
 * Protocol (OSP) server.  The Open Settlement Protocol is an ETSI
 * defined standard for Inter-Domain VoIP pricing, authorization
 * and usage exchange.  The technical specifications for OSP
 * (ETSI TS 101 321 V4.1.1) are available at www.etsi.org.
 *
 * Uli Abend was the original contributor to this module.
 *
 * Copyright (C) 2001-2005 Fhg Fokus
 *
 * This file is part of opensips, a free SIP server.
 *
 * opensips is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * opensips is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */

#ifndef _OSP_MOD_OSPTOOLKIT_H_
#define _OSP_MOD_OSPTOOLKIT_H_

#include <osp/osp.h>

/*
 * This module implements help functions for the OSP toolkit.
 * Some of the functions maybe impemented by the toolkit in the future.
 */

/*
 * Returns the OSP transaction id generated by the server.  The id is
 * returned in the Authorization Response and also packaged in the token.
 */
unsigned long long ospGetTransactionId(OSPTTRANHANDLE transaction);
void ospReportUsageWrapper(
    OSPTTRANHANDLE ospvTransaction,     /* In - Transaction handle */
    unsigned ospvReleaseCode,           /* In - Release code */
    unsigned ospvDuration,              /* In - Length of call */
    time_t ospvStartTime,               /* In - Call start time */
    time_t ospvEndTime,                 /* In - Call end time */
    time_t ospvAlertTime,               /* In - Call alert time */
    time_t ospvConnectTime,             /* In - Call connect time */
    unsigned ospvIsPDDInfoPresent,      /* In - Is PDD Info present */
    unsigned ospvPostDialDelay,         /* In - Post Dial Delay, in seconds */
    OSPE_RELEASE ospvReleaseSource      /* In - EP that released the call */
);

#endif /* _OSP_MOD_OSPTOOLKIT_H_ */

